<?php
/**
* Linkback auto-discovery code generator class file.
* @author		Chris Yates <chris.l.yates@gmail.com>
* @copyright	Copyright 2010 PBM Web Development - All Rights Reserved
* @package		linkback
*
* Copyright (c) 2011 by PBM Web Development
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
* Redistributions of source code must retain the above copyright notice, this
* list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice,
* this list of conditions and the following disclaimer in the documentation
* and/or other materials provided with the distribution.
*
* Neither the name of PBM Web Development nor the names of its contributors may
* be used to endorse or promote products derived from this software without
* specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*/
/**
* Adds {@link http://www.hixie.ch/specs/pingback/pingback Pingback}
* and {@link http://www.sixapart.com/pronet/docs/trackback_spec TrackBack}
* server auto-discovery code to posts.
*
* Useage
* ======
* Add LinkbackAutoDiscovery as a filter for the action that displays
* posts.
* <code>
* public function filters() {
*	 return array(
*		 // Other filter definitions
*		 array(
*			'path.to.linkback.LinkbackAutoDiscovery + displayActionId',
*			// Other LinkbackAutoDiscovery configuration
*		 ),
*		 // Other filter definitions
*	 );
* }
* </code>
*/
class LinkbackAutoDiscovery extends COutputProcessor {
	const PINGBACK = 'pingback';
	const TRACKBACK = 'trackback';
	const TRACKBACK_RDF = "<!--\n<rdf:RDF xmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\" xmlns:dc=\"http://purl.org/dc/elements/1.1/\" xmlns:trackback=\"http://madskills.com/public/xml/rss/module/trackback/\"><rdf:Description rdf:about=\"{permalink}\" dc:identifier=\"{permalink}\" {title}trackback:ping=\"{serverURL}\" /></rdf:RDF>\n-->\n";

	/**
	* @property string The name of the owner property that contains the model.
	* This property is only used in TrackBack auto-discovery code generation and
	* then only if url['params'] references the model or the titleAttribute is set.
	*/
	public $model = 'post';
	/**
	* @property array Routes to the local linkback servers.
	* The array keys are the server types to generate auto-discovery code for;
	* valid keys are LinkbackAutoDiscovery::PINGBACK and
	* LinkbackAutoDiscovery::TRACKBACK.
	* Array values are URL routes to the servers in the format of ControllerID/ActionID.
	* See {@link CController::createAbsoluteUrl()}
	*/
	public $servers = array(
		self::PINGBACK=>'/linkback/pingback',
		self::TRACKBACK=>'/linkback/trackback'
	);
	/**
	* @property string The model attribute that contains the document title.
	* This property is only used in TrackBack auto-discovery code generation and
	* is optional.
	*/
	public $titleAttribute;
	/**
	* @property array URL to the document in the format of key=>value pairs where the
	* keys are the parameter names of {@link http://www.yiiframework.com/doc/api/1.1/CController#createAbsoluteUrl-detail CController::createAbsoluteUrl()}
	* - route, params, schema, and ampersand, and the values the parameter values;
	* omit keys that take the default values.
	* The params value is an array of GET parameters in name=>value pairs. If the
	* value of a GET parameter starts with a dot (.) it is an attribute path
	* relative to the model; for example array('id'=>'.id') will use the model->id
	* attribute.
	* This property is only used in TrackBack auto-discovery code generation.
	* See {@link CController::createAbsoluteUrl()}
	*/
	public $url = array();

	/**
	* @var array The resolved URL
	*/
	private $_url = array(
		'route'=>'',
		'params'=>array(),
		'schema'=>'',
		'ampersand'=>'&'
	);

	/**
	* Initialises the widget.
	*/
	public function init() {
		$this->_url = array_merge($this->_url, $this->url);
		parent::init();
	}

	/**
	* Processes the captured output.
	* Adds linkback server auto-discovery code to the output.
	* @param string $output the captured output to be processed
	*/
	public function processOutput($output) {
		parent::processOutput($this->addAutoDiscovery($output));
	}

	/**
	* Adds auto-discovery code for the defined {@link servers} at the end of the
	* <head/> element.
	* If the Pingback server is specified the X-Pingback HTTP header and the
	* pingback link element are added.
	* If the TrackBack server is specified, TrackBack RDF enclosed in HTML
	* comments is added.
	* @param string The content to which auto-discovery code is added
	* @return string The content to which auto-discovery code is added
	*/
	protected function addAutoDiscovery($content) {
		if (isset($this->servers[self::TRACKBACK]))
			$this->resolveParams();
		$owner = $this->owner;
		$autoDiscover = '';
		foreach ($this->servers as $linkback=>$server) {
			$serverUrl = $owner->createAbsoluteUrl($server, array());
			switch($linkback) {
				case self::PINGBACK:
					header("X-Pingback: $serverUrl");
					$autoDiscover .= CHtml::linkTag('pingback', null, $serverUrl)."\n";
					break;
				case self::TRACKBACK:
					$url = $owner->createUrl(
						$this->_url['route'],
						$this->_url['params'],
						$this->_url['schema'],
						$this->_url['ampersand']
					);
					$autoDiscover .= strtr(self::TRACKBACK_RDF, array(
						'{serverURL}'=>$serverUrl.$url,
						'{permalink}'=>Yii::app()->getRequest()->getHostInfo().$url,
						'{title}'=>(
							empty($this->titleAttribute)?
							'':
							'dc:title="'.$owner->{$this->model}->{$this->titleAttribute}.'" '
						)
					));
					break;
			}
		} // foreach

		return preg_replace('%(</head>)%is', $autoDiscover.'\1', $content);
	}

	/**
	* Resolves dot-notation GET parameters to their values
	*/
	private function resolveParams() {
		$_model = $this->owner->{$this->model};
		foreach ($this->_url['params'] as $k=>$v) {
			if ($v[0]==='.') {
				$segs = explode('.',substr($v,1));
				$name = array_pop($segs);
				$model = $_model;
				foreach ($segs as $seg) {
					if (!array_key_exists($seg, $model->getMetaData()->relations))
						break;
					$model=$model->$seg;
				}
				$this->_url['params'][$k] = $model->$name;
			}
		}
	}
}