<?php
/**
* Linkback behavior class file.
* @author		Chris Yates <chris.l.yates@gmail.com>
* @copyright	Copyright 2010 PBM Web Development - All Rights Reserved
* @package		linkback
*
* Copyright (c) 2011 by PBM Web Development
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
* Redistributions of source code must retain the above copyright notice, this
* list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice,
* this list of conditions and the following disclaimer in the documentation
* and/or other materials provided with the distribution.
*
* Neither the name of PBM Web Development nor the names of its contributors may
* be used to endorse or promote products derived from this software without
* specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*/
/**
* {@link http://en.wikipedia.org/wiki/Linkback Linkback} Behavior class.
* Implements {@link http://www.hixie.ch/specs/pingback/pingback Pingback}
* and {@link http://www.sixapart.com/pronet/docs/trackback_spec TrackBack}
* clients.
*
* Servers are searched for in the target in the order defined by linkbacks.
* When a server is found it is pinged. The search is stopped after a successful
* response.
*
* Linkbacks are recorded in linkbacksTable to prevent duplicate pings.
*
* Useage
* ======
* Attach LinkbackClientBehavior to the posts controller and call
* LinkbackClientBehavior::linkback() from the action that displays posts.
* NOTE: The post MUST be publicly visible because the linkback server may access
* it to prevent linkback spamming.
*
	<code>
	public function behaviours() {
		return array(
			// Other behaviour declarations
			'linkbackClient'=>array(
				'class'=>'path.to.LinkbackClientBehavior',
				// LinkbackClientBehavior configuration
			),
			// Other behaviour declarations
		);
	}
	</code>

	<code>
	public function actionView() {
		$post = $this->getPost();

		if ($post->isVisible) {
			$this->linkback($post, $this->createAbsoluteUrl('', array(GET params)));
			$this->render('view', compact('post'));
		}
	}
	</code>
* @package	linkback
*/
class LinkbackClientBehavior extends CBehavior {
	const PINGBACK = 'pingback';
	const TRACKBACK = 'trackback';

	const LINKBACK_REGISTERED	= 0;
	const TARGET_NOT_FOUND	= 4;
	const SERVER_NOT_FOUND	= 8;
	const SERVER_READ_ERROR = 9;
	const LINKBACK_ALREADY_REGISTERED = 0x0030;
	const LINKBACK_INVALID_RESPONSE = 0xffff;

	const REGEX_TRACKBACK_RESPONSE = '%<response>.*?</response>%s';
	const REGEX_TRACKBACK_ERROR = '%<error>(.*?)</error>%s';
	const REGEX_TRACKBACK_MESSAGE = '%<message>(.*?)</message>%s';

	const REGEX_HTTP = '%^(http://[-A-Z0-9+&@#/\%?=~_|$!:,.;]*[A-Z0-9+&@#/\%=~_|$])$%si';
	const REGEX_HREF_HTTP = '%<a[^>]*href=["\'](http://[-A-Z0-9+&@#/\%?=~_|$!:,.;]*[A-Z0-9+&@#/\%=~_|$])["\'][^>]*>[^<]*</a>%si';
	const REGEX_PINGBACK_HEADER = '%/^X-Pingback:\s*(http:://[-A-Z0-9+&@#/\%?=~_|$!:,.;]*[A-Z0-9+&@#/\%=~_|$])$%m';
	const REGEX_PINGBACK_LINK = '%(?(?=<!--).*?-->|<link\s+rel=["\']pingback["\']\s+href=["\'](http://[-A-Z0-9+&@#/\%?=~_|$!:,.;]*[A-Z0-9+&@#/\%=~ |$])["\'](?:\s*/)?>)%si'; // This regex ensures the link tag is not inside a comment
	const REGEX_TRACKBACK_RDF = '/<rdf:RDF.*?(dc:identifier|trackback:ping)=["\'](http(?:s?):\/\/[-A-Z0-9+&@#\/%?=~ |$!:,.;]*[A-Z0-9+&@#\/%=~_|$])["\'].*?(dc:identifier|trackback:ping)=["\'](http(?:s?):\/\/[-A-Z0-9+&@#\/%?=~_|$!:,.;]*[A-Z0-9+&@#\/%=~_|$])["\'].*?<\/rdf:RDF>/si';

	const XML_HEADER = "<?xml version=\"1.0\" encoding=\"{encoding}\"?>";
	const XML_RPC_METHOD_CALL = '<methodCall><methodName>{name}</methodName>{params}></methodCall>';
	const XML_RPC_PARAMS = '<params>{params}</params>';
	const XML_RPC_PARAM = '<param>{value}</param>';
	const XML_RPC_STRING = '<value><string>{string}</string></value>';

	/**
	* @property mixed The post attributes to search for URLs.
	*
	* + string: comma delimited list of attributes
	* + array: array of attributes
	*
	* URLs found in these attributes are searched for linkback server auto-detect
	* code and a ping sent to the server if found.
	*
	* The content of these attributes must be visible in the post when it is read.
	*/
	public $attributes = 'content';
	/**
	* @property string The blog name.
	* Set empty to not send the blog name or to define the blog name when
	* generating linkbacks.
	* The blog name is only sent in TrackBack pings.
	*/
	public $blogName;
	/**
	* @property string Pingback ping encoding.
	*/
	public $encoding = 'utf-8';
	/**
	* @property string The post attribute used to extract excerpts.
	* Set to empty not to include excerpts in TrackBack pings.
	* Excerpts are only sent in TrackBack pings.
	*/
	public $excerptAttribute = 'content';
	/**
	* @property string The string used to truncate TrackBack excerpts.
	* The excerpt will be up to and including the first occurrence of the
	* delimiter. The excerpt may be further truncated to {@link excerptLength}.
	* Set empty not to truncate excerpts using a delimiter.
	* Excerpts are only sent in TrackBack pings.
	*/
	public $excerptDelimiter = '.';
	/**
	* @property string The ending added to TrackBack excerpts.
	* {@link excerptLength} includes the length of this property, i.e. the excerpt
	* text will be shortened by the length of this property.
	* Excerpts are only sent in TrackBack pings.
	*/
	public $excerptEnding = '...';
	/**
	* @property boolean Whether the length of TrackBack excerpts is exact.
	* Only valid if {@link excerptLength} is given.
	* If FALSE the excerpt is truncated on a word boundary such that the excerpt
	* length is less than or equal to {@link excerptLength}.
	* If TRUE the excerpt is truncated to exactly {@link excerptLength}.
	* Excerpts are only sent in TrackBack pings.
	*/
	public $excerptExact = false;
	/**
	* @property integer The maximum length of TrackBack excerpts.
	* Set empty not to truncate excerpts.
	* Excerpts are only sent in TrackBack pings.
	*/
	public $excerptLength = 200;
	/**
	* @property mixed The types of linkbacks we will send.
	* Either an array or a comma delimited string of linkback types. Servers are
	* searched for in the order given; the first discovered server is used.
	* Valid linkback types are 'pingback' and 'trackback'.
	*/
	public $linkbacks = 'pingback,trackback';
	/**
	* @property string Name of the tabel used to store successful linkback pings.
	* Used to prevent duplicate linkback pings being sent.
	*/
	public $linkbacksTable = 'linkbacks';
	/**
	* @property string Name of the tabel used to store save active linkback pings.
	* Used to prevent pings being sent.
	*/
	public $pingsTable = 'linkback_pings';
	/**
	* @property string The post attribute containing the post title.
	* Set empty to not send the title.
	* The post title is only sent in TrackBack pings.
	*/
	public $titleAttribute = 'title';

	/**
	* @var CDbConnection Instance CDbConnection.
	*/
	private $_connection;
	/**
	* @var array Trace messages.
	*/
	private $_traceMessages = array(
		self::LINKBACK_REGISTERED=>'Linkback to {target} registered',
		self::TARGET_NOT_FOUND=>'Target URL ({target}) not found',
		self::SERVER_NOT_FOUND=>'Could not connect to server: {server} (Target: {target})',
		self::SERVER_READ_ERROR=>'Could not read data from server: {server} (Target: {target})',
		self::LINKBACK_INVALID_RESPONSE=>'Invalid response from server: {server} (Target: {target})',
	);

	/**
	* Attach LinkbackBehavior.
	* If the tables for recording successful linkback pings and current pings do
	* not exist they are created.
	*/
	public function attach($owner) {
		parent::attach($owner);

		$this->_connection = Yii::app()->getDb();
		if (is_null($this->_connection->getSchema()->getTable($this->linkbacksTable)))
			$this->_connection->createCommand()->createTable($this->linkbacksTable, array(
				'post_id'=>'integer UNSIGNED',
				'target_url'=>'string NOT NULL',
				'PRIMARY KEY (`post_id`, `target_url`)'
			));

		if (is_null($this->_connection->getSchema()->getTable($this->pingsTable)))
			$this->_connection->createCommand()->createTable($this->pingsTable, array(
				'post_id'=>'integer UNSIGNED',
				'PRIMARY KEY (`post_id`)'
			));
	}

	/**
	* Generates linkback pings to external URLs in the post that have a linkback
	* server.
	* @param CModel The post.
	* @param string The absolute URL to the post.
	* @param string The name of the blog. If set, this overwrites the blog name
	* set in the configuration. Used in multi-blog applications where the blog is
	* determined from the post.
	*/
	public function linkback($post, $sourceUrl, $blogName=null) {
		if ($this->getIsPinging($post))
			return;
		else
			$this->setIsPinging($post);

		if (is_string($blogName))
			$this->blogName = $blogName;

		// Parse the post for external links (target URLs)
		$targetUrls = array();
		foreach ($this->toArray($this->attributes) as $attribute)
			if (isset($post->$attribute) &&
				(preg_match_all(self::REGEX_HTTP, $post->$attribute, $matches) ||
				preg_match_all(self::REGEX_HREF_HTTP, $post->$attribute, $matches)))
					$targetUrls = array_merge($targetUrls, $matches[1]);

		if (!empty($targetUrls))
			// Remove target URLs that have a linkback
			$targetUrls = array_diff($targetUrls, $this->_connection->createCommand()->select('target_url')->from($this->linkbacksTable)->where('post_id='.$post->primaryKey)->queryColumn());

		// Search for linkback servers in the target URLs
		foreach ($targetUrls as $targetUrl) {
			$params = array('http'=>array('method'=>'GET'));
			$context = stream_context_create($params);
			$fp = @fopen($targetUrl, 'rb', false, $context);

			if ($fp!==false) {
				// Send a ping to the first linkback server found, if any.
				foreach ($this->toArray($this->linkbacks) as $linkback) {
					if (method_exists($this, $linkback)) {
						if ($this->$linkback($fp, $sourceUrl, $targetUrl, $post)) {
							$this->registerLinkback($post, $targetUrl);
							$this->trace($sourceUrl, $linkback, self::LINKBACK_REGISTERED, compact('targetUrl'));
							break;
						}
					}
					else {
						$this->setIsPinging($post, false);
						throw new CException(Yii::t('linkback',
								'Unsupported linkback type: {linkback}',
								array('{linkback}'=>$linkback)));
					}
				}
				fclose($fp);
			}
			else
				$this->trace($sourceUrl, 'linkback', self::TARGET_NOT_FOUND, compact('targetUrl'));
		}

		$this->setIsPinging($post, false);
	}

	/**
	* Returns the isPinging status of the post.
	* @param CModel The post.
	* @return boolean True is the post is alrady generating pings, false if not.
	*/
	private function getIsPinging($post) {
		return $this->_connection->createCommand()->select('COUNT(*)')->from($this->pingsTable)->where('post_id='.$post->primaryKey)->queryScalar()>0;
	}

	/**
	* Sets the isPinging status of the post.
	* @param CModel The Post.
	* @param boolean The isPinging status.
	*/
	private function setIsPinging($post, $isPinging=true) {
		if ($isPinging)
			$this->_connection->createCommand()->insert($this->pingsTable, array('post_id'=>$post->primaryKey));
		else
			$this->_connection->createCommand()->delete($this->pingsTable, 'post_id='.$post->primaryKey);
	}

	/**
	* Registers a linkback from a post to the target URL.
	* @param string The target URL of the linkback ping.
	* @param Cmodel The post.
	* @return boolean TRUE if the registration was successful, FALSE if not.
	*/
	private function registerLinkback($post, $targetUrl) {
		if ($this->_connection->createCommand()->insert($this->linkbacksTable, array(
			'post_id'=>$post->primaryKey,
			'target_url'=>$targetUrl,
		)) > 0) {
			$this->linkbackRegistered($post, $targetUrl);
			return true;
		}
		return false;
	}

	/**
	 * This method is invoked after a linkback has been successfully registered.
	 * The default implementation raises the {@link onLinkbackRegistered} event.
	 * Override this method to do postprocessing after linkback registration.
	 * Make sure you call the parent implementation so that the event is raised properly.
	 * @param CModel The source post
	 * @param string The target URL
	 */
	protected function linkbackRegistered($post, $targetUrl) {
		if ($this->hasEventHandler('onLinkbackRegistered'))
			$this->onLinkbackRegistered(new CEvent($this, array(
				'name'=>'onLinkbackRegistered',
				'post'=>$post,
				'targetUrl'=>$targetUrl
			)));
	}

	/**
	 * This event is raised after a linkback has been successfully registered.
	 * @param CEvent The event parameter
	 */
	public function onLinkbackRegistered($event)	{
		$this->raiseEvent('onLinkbackRegistered',$event);
	}

	/**
	* Send trace message.
	* @param string The source URL
	* @param string The type of linkback
	* @param integer Error code
	* @param mixed string: An error message; array: Parameters for a predefined error message
	*/
	private function trace($source, $type, $code, $message) {
		if (is_array($message))
			$message = Yii::t('linkback', $this->_traceMessages[$code], array(
				'{server}'=>(isset($message['server'])?$message['server']:''),
				'{target}'=>(isset($message['targetUrl'])?$message['targetUrl']:'')
			));
		Yii::trace("{$source}: $type: Code $code: $message", 'linkback.LinkbackClientBehavior');
	}

	/*****************
	* Pingback Client
	******************/

	/**
	* Send a Pingback ping.
	* @param resource File pointer to the target URL
	* @param string The source URL
	* @return boolean True if the Pingback completed successfully, false if not.
	*/
	private function pingback($fp, $sourceUrl, $targetUrl) {
		$server = $this->findPingbackServer($fp);
		if (empty($server))
			return false;

		$xmlrpcString = str_replace('{string}', $sourceUrl, self::XML_RPC_STRING);
		$source = str_replace('{value}', $xmlrpcString, self::XML_RPC_PARAM);
		$xmlrpcString = str_replace('{string}', $targetUrl, self::XML_RPC_STRING);
		$target = str_replace('{value}', $xmlrpcString, self::XML_RPC_PARAM);
		$xmlrpcParams = str_replace('{params}', $source.$target, self::XML_RPC_PARAMS);
		$xmlrpcMethodCall = strtr(self::XML_RPC_METHOD_CALL, array(
			'{name}'=>'pingback.ping',
			'{params}'=>$xmlrpcParams
		));
		$pingback = str_replace('{encoding}', $this->encoding, self::XML_HEADER) .
			$xmlrpcMethodCall;

		$params = array('http'=>array(
			'method'=>'POST',
			'ignore_errors'=>true,
			'header'=>"Content-Type: text/xml\nContent-Length: ". strlen($pingback),
			'content'=>$pingback
		));
		$context = stream_context_create($params);
		$serverFp = @fopen($server, 'rb', false, $context);
		if (!$serverFp) {
			$this->trace($sourceUrl, self::PINGBACK, self::SERVER_NOT_FOUND, compact('server', 'targetUrl'));
			return false;
		}

		$response = @simplexml_load_string(@stream_get_contents($serverFp));
		fclose($serverFp);

		if (empty($response)) {
			$this->trace($sourceUrl, self::PINGBACK, self::SERVER_READ_ERROR, compact('server', 'targetUrl'));
			return false;
		}

		if ($response->params)
			return true;
		if ($response->fault) {
			$code = (string)$response->fault->value->struct->member[0]->value->int;
			$message = (string)$response->fault->value->struct->member[1]->value->string;
		}
		else {
			$code = self::LINKBACK_INVALID_RESPONSE;
			$message = compact('server', 'targetUrl');
		}
		$this->trace($sourceUrl, self::TRACKBACK, $code, $message);
		if ($code===self::LINKBACK_ALREADY_REGISTERED)
				return true;
		return false;
	}

	/**
	* Returns the Pingback server URL if either the X-Pingback header or the
	* <link rel="pingback" /> tag if present.
	* @param resource File pointer to the target.
	* @return string The Pingback server URL.
	*/
	private function findPingbackServer($fp) {
		$server = $this->findPingbackServerInHeader($fp);
		return (is_null($server))?$this->findPingbackServerInLink($fp):$server;
	}

	/**
	* Returns the Pingback server URL from the X-Pingback header if present.
	* @param resource File pointer
	* @return string The Pingback server URL if the X-Pingback header is present,
	* else NULL.
	*/
	private function findPingbackServerInHeader($fp) {
		$headers = stream_get_meta_data($fp);
		foreach ($headers['wrapper_data'] as $http_header)
			if (preg_match(self::REGEX_PINGBACK_HEADER, $http_header, $matches))
				return $matches[1];
	}

	/**
	* Returns the Pingback server URL from the <link rel="pingback" /> tag if
	* present in the <head/> tag.
	* @param resource file pointer
	* @return string The Pingback server URL if the <link rel="pingback" /> tag is
	* present in the <head/> tag, else NULL.
	*/
	private function findPingbackServerInLink($fp) {
		do {
			$line = fgets($fp);
			if (preg_match(self::REGEX_PINGBACK_LINK, $line, $matches) && isset($matches[1]))
				return $matches[1];
		} while (strpos($line, '</head>')===false);
	}

	/*****************
	* TrackBack Client
	******************/

	/**
	* Send a TrackBack ping.
	* @param resource File pointer to the target URL.
	* @param string The source URL.
	* @param string The target URL of the linkback ping.
	* @param Cmodel The post.
	* @return boolean True if the TrackBack ping was successfull, false if not.
	*/
	private function trackback($fp, $sourceUrl, $targetUrl, $post) {
		$server = $this->findTrackbackServer($fp, $targetUrl);
		if (is_null($server))
			return false;

		$trackback = array();
		$trackback['url'] = $sourceUrl;
		if (!empty($this->blogName))
			$trackback['blog_name'] = $this->blogName;
		if (!empty($post->{$this->titleAttribute}))
			$trackback['title'] = $post->{$this->titleAttribute};
		if (!empty($post->{$this->excerptAttribute}))
			$trackback['excerpt'] = $this->excerpt($post->{$this->excerptAttribute});
		$params = array('http'=>array(
			'method'=>'POST',
			'ignore_errors'=>true,
			'content'=>http_build_query($trackback)
		));
		$context = stream_context_create($params);
		$serverFp = @fopen($server, 'rb', false, $context);
		if (!$serverFp) {
			$this->trace($sourceUrl, self::TRACKBACK, self::SERVER_NOT_FOUND, compact('server', 'targetUrl'));
			return false;
		}

		$response = @stream_get_contents($serverFp);
		fclose($serverFp);

		if (empty($response)) {
			$this->trace($sourceUrl, self::TRACKBACK, self::SERVER_READ_ERROR, compact('server', 'targetUrl'));
			return false;
		}

		if (preg_match(self::REGEX_TRACKBACK_RESPONSE, $response)) {
			preg_match(self::REGEX_TRACKBACK_ERROR, $response, $code);
			$code = trim($code[1]);
			if ($code==='0')
				return true;
			else {
				preg_match(self::REGEX_TRACKBACK_MESSAGE, $response, $message);
				if (isset($message[1]))
					$message = trim($message[1]);
				else {
					$code = self::LINKBACK_INVALID_RESPONSE;
					$message = compact('server', 'targetUrl');
				}
			}
		}
		else {
			$code = self::LINKBACK_INVALID_RESPONSE;
			$message = compact('server', 'targetUrl');
		}
		$this->trace($sourceUrl, self::TRACKBACK, $code, $message);
		if ($code==self::LINKBACK_ALREADY_REGISTERED)
				return true;
		return false;
	}

	/**
	* Returns the target TrackBack server URL if TrackBack auto-discovery RDF is
	* present in the <head/> tag.
	* @param resource File pointer to the target.
	* @param string The target URL of the linkback ping.
	* @return string The target TrackBack server URL if TrackBack auto-discovery
	* RDF is present in the <head/> tag, else NULL.
	*/
	private function findTrackbackServer($fp, $targetUrl) {
		$file = stream_get_contents($fp);
		preg_match_all(self::REGEX_TRACKBACK_RDF, $file, $matches, PREG_SET_ORDER);
		foreach ($matches as $match) {
			if ($match[1]==='dc:identifier' && $match[2]===$targetUrl)
				return $match[4];
			if ($match[3]==='dc:identifier' && $match[4]===$targetUrl)
				return $match[2];
		} // foreach
	}

	/**
	* Returns an excerpt from the content.
	* All tags are stripped from the content.
	* @param string The content
	* @return string The excerpt
	*/
	private function excerpt($content) {
		$excerpt = strip_tags($content);
		if (!empty($this->excerptDelimiter))
			$excerpt = substr($excerpt, 0, strpos($excerpt, $this->excerptDelimiter) + strlen($this->excerptDelimiter));
		if (!empty($this->excerptLength)) {
			$excerpt = substr($excerpt, 0, $this->excerptLength - strlen($this->excerptEnding));
			if (!$this->excerptExact) {
				$pos = strrpos($excerpt,' ');
				if ($pos!==false)
					$excerpt = substr($excerpt, 0 ,$pos);
			}
		}
		return $excerpt.(strlen($excerpt)<strlen($content)?$this->excerptEnding:'');
	}

	/**
	* Converts a delimited list into an array.
	* If the list is an array it is returned unaltered.
	* @param mixed The list
	* @param string The list delimiter
	* @return array The list converted to an array
	*/
	private function toArray($list,$delimiter=',') {
		return (is_string($list)?array_map('trim',explode($delimiter,$list)):$list);
	}
}