<?php
/**
* LinkbackController class file.
* @author		Chris Yates <chris.l.yates@gmail.com>
* @copyright	Copyright 2010 PBM Web Development - All Rights Reserved
* @package		linkback
*
* Copyright (c) 2010 by PBM Web Development
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
* Redistributions of source code must retain the above copyright notice, this
* list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice,
* this list of conditions and the following disclaimer in the documentation
* and/or other materials provided with the distribution.
*
* Neither the name of PBM Web Development nor the names of its contributors may
* be used to endorse or promote products derived from this software without
* specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*/
/**
* LinkbackController class.
* Implements {@link http://www.hixie.ch/specs/pingback/pingback Pingback}
* and {@link http://www.sixapart.com/pronet/docs/trackback_spec TrackBack}
* {@link http://en.wikipedia.org/wiki/Linkback Linkback} servers.
*
* Accepted pings to the servers are stored in the feedback table (usually also
* stores comments).
*
* For a ping to be accepted as a valid linkback it must:
* + Come from a source URL that can be read and has a vaild content-type
* + Be to a valid target URL, i.e. one that points to a URL that accepts
* linkbacks
* + The source must contain a link to the target URL
* + Not be previously registered
*
* Server responses are logged.
*
* Support is provided for two schemas:
* <pre>
* 1. Single blog with feedback (comments, pingbacks, and trackbacks)
* -------------		-----------------
* | tbl_posts |		| tbl_feedback	|
* |-----------|		|---------------|
* | PK | id	 |<-- | PK	| id			|
* |-----------|	| |-----|---------|
* |		|			|	--| FK1 | post_id |
* -------------		-----------------
*
* 2. Multiple blogs
* -------------		-----------------		-----------------
* | tbl_blogs |		| tbl_posts		 |		| tbl_feedback	|
* |-----------|		|---------------|		|---------------|
* | PK | id	 |<-- | PK	| id			|<-- | PK	| id			|
* |-----------|	| |---------------|	| |-----|---------|
* |		|			|	--| FK1 | blog_id |	--| FK1 | post_id |
* -------------		-----------------		-----------------
* </pre>
*
* Useage
* ======
* Set controllerMap in your config:
* <code>
		'controllerMap'=>array(
			'linkback'=>array(
				'class'=>'path.to.linkback.extension.controllers.linkbackController',
				'uriRegex'=>'(myblog)/post/([a-z0-9]+(-[a-z0-9]+)*)'
			)
		)
* </code>
*
* Note that pages can have multiple instances of embedded RDF; clients should
* use the block whose dc:identifier matches the target URL they wish to ping.
*
* <b>WARNING:</b>
* If you have debug turned on that outputs to the browser, server responses
* may be seen by clients as non-well formed XML due to the extra content.
*
* @package			linkback
*/
class LinkbackController extends CExtController {
	const PINGBACK = 'pingback';
	const TRACKBACK = 'trackback';

	const REGEX_CONTENT_TYPE = '%Content-Type:(?:\s+.*?)*\s+(.*?/.*?);%';
	const LINKBACK_REGISTERED = 0;
	/** Error codes defined in {@link http://www.hixie.ch/specs/pingback/pingback#TOC3 the pingback specification}*/
	const SOURCE_URI_DOES_NOT_EXIST = 0x0010;
	const SOURCE_URI_DOES_NOT_LINK_TO_TARGET_URI = 0x0011;
	const TARGET_URI_DOES_NOT_EXIST = 0x0020;
	const TARGET_URI_CANNOT_BE_USED = 0x0021;
	const LINKBACK_ALREADY_REGISTERED = 0x0030;
	const ACCESS_DENIED = 0x0031;
	/** Additional error codes */
	const SOURCE_URI_NOT_SPECIFIED = 0x0012;
	const LINKBACK_NOT_REGISTERED = 0x0032;
	const INVALID_CONTENT_TYPE = 0x0033;
	/** XML-RPC error codes defined in {@link http://xmlrpc-epi.sourceforge.net/specs/rfc.fault_codes.php the Specification for Fault Code Interoperability, version 20010516}*/
	const NOT_WELL_FORMED = -32700;
	const INVALID_XML_RPC = -32600;
	const APPLICATION_ERROR = -32500;
	const REQUESTED_METHOD_NOT_FOUND = -32601;
	const INVALID_METHOD_PARAMETERS = -32602;
	const APPLICATION_ERROR_TEXT = 'Application Error';
	const XML_HEADER = "<?xml version=\"1.0\" encoding=\"{encoding}\"?>";
	const XML_RPC_METHOD_RESPONSE = '<methodResponse>{response}></methodResponse>';
	const XML_RPC_PARAMS = '<params>{params}</params>';
	const XML_RPC_PARAM = '<param>{value}</param>';
	const XML_RPC_FAULT = '<fault><value><struct>{members}</struct></value></fault>';
	const XML_RPC_MEMBER = '<member><name>{name}</name>{value}</member>';
	const XML_RPC_STRING = '<value><string>{string}</string></value>';
	const XML_RPC_INT = '<value><int>{int}</int></value>';
	const TRACKBACK_RESPONSE = '<response><error>{code}</error><message>{msg}</message></response>';

	/**
	* @property array Maps LinkbackController internal properties to linkback model attributes.
	* Each entry maps an internal property to a linkback model attribute.	Omit
	* any non-required properties that do not have an equivalent attribute in
	* your linkback model.
	*
	* LinkbackController internal properties	are:
	* + postFk=>the post foriegn key (required)
	* + sourceUrl=>the source URL (required)
	* + type=>type of linkback; pingback or trackback (required)
	* + blogName=>the source blog name (Trackbacks only and must be in the
	* ping)
	* + content=>the "comment" about the linkback. $content defines the template.
	* + status=>status of this record on save. Defined by $status.
	* + ipAddress=>source ip_address
	* + title=>title of the source post
	* Leave out any not used by your feedback model.
	*/
	public $attributeMap = array(
		'sourceUrl'	=>'url',
		'targetId'	=>'post_id',
		'type'			=>'type',
		'blogName'	=>'author',
		'content'		=>'content',
		'ipAddress'	=>'ip_address',
		'status'		=>'status',
		'title'			=>'title'
	);
	/**
	* @property string Template for content to be stored when a linkback is
	* registered.
	*
	* The following placeholders are recognised and will be replaced:
	* + {excerpt}=>an excerpt from the source (Trackbacks only and must be in
	* the ping)
	* + {blogName}=>the source blog name (Trackbacks only and must be in the
	* ping)
	* + {sourceUrl}=>the source URL
	* + {title}=>either the title of the source or the source URL if the title
	* is not given.
	* + {type}=>the type of linkback; Pingback or Trackback.
	*
	* To make a string containing a placeholder optional (i.e. rendered only if
	* the placeholder has a value) enclose it in square brackets. Escape square
	* brackets that are required in the output with a backslash.
	*
	* Example: "{type} from {title}[\n\n{excerpt}]" is the equivalent of
	* '{type} from {title}' if excerpt is empty.
	* @see LinkbackController::attributeMap
	*/
	public $content = "{type} from {title}[\n\n{excerpt}]";
	/**
	* @property array Content types from which linkbacks are accepted.
	*/
	public $contentTypes = array('text/html', 'text/xml', 'application/xhtml+xml');
	/**
	* @property string Response encoding.
	*/
	public $encoding = 'utf-8';
	/**
	* @property string The ending added to TrackBack excerpts.
	* {@link excerptLength} includes the length of this property, i.e. the excerpt
	* text will be shortened by the length of this property.
	* Excerpts are only sent in TrackBack pings.
	*/
	public $excerptEnding = '...';
	/**
	* @property boolean Whether the length of TrackBack excerpts is exact.
	* Only valid if {@link excerptLength} is given.
	* If FALSE the excerpt is truncated on a word boundary such that the excerpt
	* length is less than or equal to {@link excerptLength}.
	* If TRUE the excerpt is truncated to exactly {@link excerptLength}.
	* Excerpts are only sent in TrackBack pings.
	*/
	public $excerptExact = false;
	/**
	* @property integer The maximum length of TrackBack excerpts.
	* The excerpt is truncated to the exerptDelimiter string forst, then to this
	* length.
	* Set to empty not to truncate excerpts.
	*/
	public $excerptLength = 255;
	/**
	* @property mixed Paths to import.
	* Use to import model inheritancies for {@link $post} and {@link linkback}
	* models if required. Leave empty if the models directly extend CActiveRecord
	* or import required inheritances themselves.
	* An array or comma delimited string of path aliases.
	*/
	public $importPaths;
	/**
	* @property string Path to the model that stores linkbacks. If the model is
	* imported in {@link importPaths} this property need only be the model class
	* name.
	*/
	public $linkback;
	/**
	* @property array Maps target URL parameters to post model attributes.
	* The format is 'URL param'=>'post model attribute'.	Related model attributes
	* are specified using dot notation; e.g. blog.id is the 'id' attribute of the
	* model specified in the 'blog' relationship of the 'post' model.
	*/
	public $params;
	/**
	* @property array Path to the model that stores posts. If the model is
	* imported in {@link importPaths} this property need only be the model class
	* name.
	*/
	public $post;
	/**
	* @property array Target URL routing rules (pattern=>route).
	*/
	public $rules;
	/**
	* @var string The linkback model scenario when a linkback is registered.
	* Use this to ensure the linkback is not validated against certain linkback
	* model validation rules by setting the value of the rule's 'on' property.
	* For example, if the linkback model is also used for user comments you may
	* have CAPTCHA validation; a linkback should not be validated against such a
	* rule.
	*/
	public $scenario = 'linkback';
	/**
	* @property mixed The value given to the linkback model's status attribute.
	* If status is specified in attributeMap, this value will be stored in the
	* linkback model's status attribute and should be set to a value that
	* indicated the linkback is approved and/or can be displayed.
	*/
	public $status = 'approved';
	/**
	* @property string Template for the title to be stored when a linkback is
	* registered.
	*
	* The following placeholders are recognised and will be replaced:
	* + {blogName}=>the source blog name (Trackbacks only and if in the ping)
	* + {sourceUrl}=>the source URL
	* + {title}=>either the title of the source or the source URL if the title
	* is not given.
	*
	* To make a string containing a placeholder optional (i.e. rendered only if
	* the placeholder has a value) enclose it in square brackets. Escape square
	* brackets that are required in the output with a backslash.
	*
	* Example: "{title}[@{blogName}]" is the equivalent of '{title}' if
	* blogName is empty.
	*/
	public $title = '{title}[@{blogName}]';
	/**
	* @property boolean Whether validation is performed when the linkback is
	* registered.
	*/
	public $validate = true;

	/**
	* @var string Type of linkback.
	*/
	private $_type;
	/**
	* @var array HTTP response header from the source. Used to check for a valid
	* content type.
	*/
	private $_http_response_header;
	/**
	* @var string Title of the source.
	*/
	private $_title;
	/**
	* @var string Excerpt of the source (Trackbacks only).
	*/
	private $_excerpt;
	/**
	* @var string Source blog name (Trackbacks only).
	*/
	private $_blogName;
	/**
	* @var string Local blog identifier (Multi-blog applications only).
	* Extracted fom target URL by uriRegex.
	*/
	private $_blog;

	/**
	* @var string Local post identifier.
	* Extracted fom target URL by uriRegex.
	*/
	private $_post;
	/**
	* @var CActiveRecord The model currently being used by the server.
	*/
	private $_linkback;
	/**
	* @var string Source URL.
	*/
	private $_sourceUrl;
	/**
	* @var string Target URL.
	*/
	private $_targetUrl;
	/**
	* @var array Response messages.
	*/
	private $_responses = array(
		self::APPLICATION_ERROR=>'{message}',
		self::LINKBACK_REGISTERED=>'{linkback} from {source} to {target} registered.',
		self::SOURCE_URI_NOT_SPECIFIED=>'Source URI not specified (target URI = {target}).',
		self::SOURCE_URI_DOES_NOT_EXIST=>'Source URI ({source}) does not exist (target URI = {target}).',
		self::SOURCE_URI_DOES_NOT_LINK_TO_TARGET_URI=>'Source URI ({source}) does not contain a link to the target URI ({target}).',
		self::TARGET_URI_DOES_NOT_EXIST=>'Target URI ({target}) does not exist (source URI = {source}).',
		self::TARGET_URI_CANNOT_BE_USED=>'Target URI ({target}) not valid (source URI = {source}).',
		self::LINKBACK_ALREADY_REGISTERED=>'{linkback} from {source} to {target} already registered.',
		self::NOT_WELL_FORMED=>'parse error. Not well formed.',
		self::INVALID_XML_RPC=>'server error. Invalid xml-rpc. Not conforming to spec.',
		self::REQUESTED_METHOD_NOT_FOUND=>'server error. Requested method not found.',
		self::INVALID_METHOD_PARAMETERS=>'server error. Invalid method parameters.',
		self::LINKBACK_NOT_REGISTERED=>'system error: {linkback} from {source} to {target} not registered.',
		self::INVALID_CONTENT_TYPE=>'transport error: Source URI ({source}) Content-Type: {contentType} is invalid.',
	);

	/**
	* Initializes the controller.
	*/
	public function init() {
		$this->_type = (preg_match('/'.self::PINGBACK.'/',Yii::app()->getRequest()->getPathInfo())?self::PINGBACK:self::TRACKBACK);
		try {
			parent::init();
			if (!empty($this->importPaths))
				foreach ($this->toArray($this->importPaths) as $path)
					Yii::import($path);

			foreach (array('linkback','params','post','rules') as $property)
				if (empty($this->$property))
					throw new CException(Yii::t('linkback', 'Linkback server property "{property}" not initialised', array('{property}'=>$property)));

			Yii::import($this->post);
			Yii::import($this->linkback);

			$post = array_pop(explode('.',$this->post));
			$linkback = array_pop(explode('.',$this->linkback));
			$this->_post = new $post;
			$this->_linkback = new $linkback;
		}
		catch (CException $e) {
			$this->response(self::APPLICATION_ERROR, array('{message}'=>(YII_DEBUG?$e->getMessage():self::APPLICATION_ERROR_TEXT)));
		}
	}

	/**
	* PingBack xmlrpc server.
	* The path to this action must go in the pingback link in the head section,
	* e.g. <link rel="pingback" href="http://example.com/posts/linkback/pingback" />,
	* or in the X-pingback header of the page accepting pingbacks.
	*/
	public function actionPingback() {
		$xmlrpc = simplexml_load_string(file_get_contents('php://input'));
		if (!$xmlrpc) {
			$this->response(self::NOT_WELL_FORMED);
		}
		if (!$xmlrpc->methodName) {
			$this->response(self::INVALID_XML_RPC);
		}
		if ((string)$xmlrpc->methodName !== 'pingback.ping') {
			$this->response(self::REQUESTED_METHOD_NOT_FOUND);
		}
		if (count($xmlrpc->params->param) != 2) {
			$this->response(self::INVALID_METHOD_PARAMETERS);
		}
		$this->_sourceUrl = (string)$xmlrpc->params->param[0]->value->string;
		if (!$this->_sourceUrl) {
			$this->response(self::INVALID_METHOD_PARAMETERS);
		}
		$this->_targetUrl = (string)$xmlrpc->params->param[1]->value->string;
		if (!$this->_targetUrl) {
			$this->response(self::INVALID_METHOD_PARAMETERS);
		}
		return $this->linkback();
	}

	/**
	* TrackBack server.
	* The path to this action is the trackback:ping attribute in the
	* autodiscovery RDF of the page accepting trackbacks.
	* e.g. trackback:ping="http://example.com/linkback/trackback/<targetURI>"
	*/
	public function actionTrackback() {
		if (!isset($_POST['url']))
			$this->response(self::SOURCE_NOT_SPECIFIED, compact('target'));

		$request = Yii::app()->getRequest();
		$this->_sourceUrl = $_POST['url'];
		$this->_targetUrl = $request->getHostInfo().substr($request->getUrl(),strlen($this->createUrl('')));

		$this->_title = (isset($_POST['title']) ? strip_tags($_POST['title']) : '');
		$this->_blogName = (isset($_POST['blog_name']) ? strip_tags($_POST['blog_name']) : '');
		$this->_excerpt = (isset($_POST['excerpt']) ? $this->excerpt($_POST['excerpt']) : '');

		$this->linkback();
	}

	/**
	* Returns an excerpt from the content.
	* All tags are stripped from the content.
	* @param string The content
	* @return string The excerpt
	*/
	private function excerpt($content) {
		$excerpt = strip_tags($content);
		if (!empty($this->excerptDelimiter))
			$excerpt = substr($excerpt, 0, strpos($excerpt, $this->excerptDelimiter) + strlen($this->excerptDelimiter));
		if (!empty($this->excerptLength)) {
			$excerpt = substr($excerpt, 0, $this->excerptLength - strlen($this->excerptEnding));
			if (!$this->excerptExact) {
				$pos = strrpos($excerpt,' ');
				if ($pos!==false)
					$excerpt = substr($excerpt, 0 ,$pos);
			}
		}
		return $excerpt.(strlen($excerpt)<strlen($content) && substr($excerpt, strlen($this->excerptEnding)*-1)!==$this->excerptEnding?$this->excerptEnding:'');
	}

	/**
	* Process a linkback ping.
	*/
	private function linkback() {
		try {
			$target = $this->getTarget();

			if (is_integer($target))	// getTarget() returned an error code
				$this->response($target,
						array('{source}'=>$this->_sourceUrl, '{target}'=>$this->_targetUrl));

			if ($this->isLinkbackRegistered($target))
				$this->response(self::LINKBACK_ALREADY_REGISTERED, array(
					'{source}'=>$this->_sourceUrl,
					'{target}'=>$this->_targetUrl,
					'{linkback}'=>'Linkback'
				));

			$fp = $this->openSource($this->_sourceUrl);
			if (!$fp)
				$this->response(self::SOURCE_URI_DOES_NOT_EXIST,
						array('{source}'=>$this->_sourceUrl, '{target}'=>$this->_targetUrl));

			$contentType = $this->validContentType();
			if ($contentType===true) {
				$sourceHasLinkToTarget = $this->sourceHasLinkToTarget($fp);
				fclose($fp);
			}
			else {
				fclose($fp);
				$this->response(self::INVALID_CONTENT_TYPE,
						array('{source}'=>$this->_sourceUrl, '{contentType}'=>$contentType));
			}

			if (!$sourceHasLinkToTarget)
				$this->response(self::SOURCE_URI_DOES_NOT_LINK_TO_TARGET_URI,
						array('{source}'=>$this->_sourceUrl, '{target}'=>$this->_targetUrl));

			if (!$this->registerLinkback($target))
				$this->response(self::LINKBACK_NOT_REGISTERED, array(
					'{source}'=>$this->_sourceUrl,
					'{target}'=>$this->_targetUrl,
					'{linkback}'=>ucfirst($this->_type)
				));
			else
				$this->response(self::LINKBACK_REGISTERED, array(
					'{source}'=>$this->_sourceUrl,
					'{target}'=>$this->_targetUrl,
					'{linkback}'=>ucfirst($this->_type)
				));
		}
		catch (CException $e) {
			$this->response(self::APPLICATION_ERROR, array('{message}'=>(YII_DEBUG?$e->getMessage():self::APPLICATION_ERROR_TEXT)));
		}
	}

	/**
	* Returns the target document model or a value indicating why it could not be.
	* If the target URL does not parse self::TARGET_URI_CANNOT_BE_USED is
	* returned.
	*
	* If the target document cannot be found self::TARGET_URI_DOES_NOT_EXIST is
	* returned.
	* @return mixed CModel: The target document model; integer: error status
	*/
	private function getTarget() {
		$pathInfo = $this->getPathInfo();
		$connection = $this->_post->getDbConnection();
		foreach ($this->rules as $pattern=>$route) {
			$rule = new CUrlRule($route, $pattern);
			if ($r = $rule->parseUrl(Yii::app()->getUrlManager(), Yii::app()->getRequest(), $pathInfo, $pathInfo)!==false) {
				$criteria = new CDbCriteria();
				$condition = array();
				foreach ($this->params as $param=>$attribute) {
					if (strpos($attribute,'.')) {
						$criteria->mergeWith($this->relatedAttributeCriteria($attribute,$param));
					}
					else
						$criteria->addCondition($this->_post->getTableAlias(true).'.'.$connection->quoteColumnName($attribute).'="'.$_GET[$param].'"');
				}

				$target = $this->_post->find($criteria);
				return (is_null($target)?self::TARGET_URI_DOES_NOT_EXIST:$target);
			}
		}
		return self::TARGET_URI_CANNOT_BE_USED;
	}

	private function getPathInfo() {
		$request = Yii::app()->getRequest();
		$hostInfo = $request->getHostInfo();
		$pathInfo=substr($this->_targetUrl, strlen($hostInfo));

		if(($pos=strpos($pathInfo,'?'))!==false)
			 $pathInfo=substr($pathInfo,0,$pos);

		$pathInfo=urldecode($pathInfo);

		$scriptUrl=$request->getScriptUrl();
		$baseUrl=$request->getBaseUrl();
		if (strpos($pathInfo,$scriptUrl)===0)
			$pathInfo=substr($pathInfo,strlen($scriptUrl));
		else if($baseUrl==='' || strpos($pathInfo,$baseUrl)===0)
			$pathInfo=substr($pathInfo,strlen($baseUrl));
		else if(strpos($_SERVER['PHP_SELF'],$scriptUrl)===0)
			$pathInfo=substr($_SERVER['PHP_SELF'],strlen($scriptUrl));
		else
			throw new CException(Yii::t('linkback','Unable to determine path info of target URL.'));

		return trim($pathInfo,'/');
	}

	/**
	* Returns a CDbCriteria object for a related model attribute
	* @param string The attribute path in dot notation
	* @param string GET parameter associated with the attribute
	* @return CDbCriteria CDbCriteria object
	*/
	private function relatedAttributeCriteria($attribute,$param) {
		$criteria = new CDbCriteria();
		$attribute = explode('.',$attribute);
		$relations = $this->_post->relations();

		if (!isset($relations[$attribute[0]]))
			throw new CException(Yii::t('linkback','Undefined relationship: {post}->{rel}', array('{post}'=>get_class($this->_post), '{rel}'=>$attribute[0])));

		$connection = CActiveRecord::model($relations[$attribute[0]][1])->getDbConnection();
		$criteria->with = $attribute[0];
		$criteria->condition = $connection->quoteTableName($attribute[0]).'.'.$connection->quoteColumnName($attribute[1]).'="'.$_GET[$param].'"';
		return $criteria;
	}

	/**
	* Determine if there is an existing linkback from the source to the target.
	* @param mixed The target primary key
	* @return boolean True if there is an existing linkback from the source to
	* the target, else false
	*/
	private function isLinkbackRegistered($target) {
		$criteria = new CDbCriteria();
		$criteria->condition = "`{$this->attributeMap['targetId']}`='{$target->primaryKey}' AND `{$this->attributeMap['sourceUrl']}`='{$this->_sourceUrl}' AND (`{$this->attributeMap['type']}`='".self::PINGBACK."' OR `{$this->attributeMap['type']}`='".self::TRACKBACK."')";

		return (boolean)$this->_linkback->count($criteria);
	}

	/**
	* Open the surce URL.
	* @return resource File pointer to source
	*/
	private function openSource() {
		$params = array(
			'http'=>array('method'=>'GET')
		);
		$context = stream_context_create($params);

		$fp = @fopen($this->_sourceUrl, 'rb', false, $context);

		if ($fp)
			$this->_http_response_header = $http_response_header;

		return $fp;
	}

	/**
	* Checks for a valid content type in the HTTP response header from the source.
	* @return mixed boolean: true if valid content type, or
	* string: invalid content type or empty string if no content type given
	*/
	private function validContentType() {
		$count = count($this->_http_response_header);
		for ($i = 0; $i < $count; $i++) {
			preg_match(self::REGEX_CONTENT_TYPE, $this->_http_response_header[$i], $matches);
			if (!empty($matches)) {
				return in_array($matches[1], $this->contentTypes) ? true : $matches[1];
			}
		}
		return '';
	}

	/**
	* Check the source contains a link to the target.
	* Also sets the title for pingbacks.
	* @param resource File pointer to the source
	* @return boolean True if the source contains a link to the target, else false
	*/
	private function sourceHasLinkToTarget($fp) {
		$file = stream_get_contents($fp);
		$sourceHasLinkToTarget =
				strstr($file, preg_replace("/&(amp;)?/i","&amp;",$this->_targetUrl)) ||
				strstr($file, str_replace("&amp;","&",$this->_targetUrl));

		if ($sourceHasLinkToTarget && $this->_type === self::PINGBACK) {
			$title = (preg_match("%<title>([^<]*)</title>%i",$file,$matches) ?
					trim($matches[1]) : '');
		}
		return $sourceHasLinkToTarget;
	}

	/**
	* Register the linkback.
	* @param CModel The target
	* @return boolean True if the linkback is registered successfuly, else false
	*/
	private function registerLinkback($target) {
		$this->_linkback->setScenario($this->scenario);

		foreach ($this->attributeMap as $parameter=>$attribute) {
			switch ($parameter) {
				case 'targetId':
					$value = $target->getPrimaryKey();
					break;
				case 'type':
					$value = $this->_type;
					break;
				case 'title':
					$value = $this->parseTemplate('title');
					break;
				case 'content':
					$value = $this->parseTemplate('content');
					break;
				case 'blogName':
					$value = ($this->_blogName ? $this->_blogName : '');
					break;
				case 'sourceUrl':
					$value = $this->_sourceUrl;
					break;
				case 'status':
					$value = $this->status;
					break;
				case 'ipAddress':
					$value = Yii::app()->request->userHostAddress;
					break;
			}
			$this->_linkback->$attribute = $value;
		} // foreach

		if ($this->_linkback->save($this->validate)) {
			$this->linkbackRegistered($target);
			return true;
		}
		return false;
	}

	/**
	* Parse a template.
	* Replace placeholders in the template with their values and remove any
	* sections where the placeholder value is empty.
	* @param string Name of the template
	* @return string The parsed template
	*/
	private function parseTemplate($template) {
		if ($template === 'content') {
			$replacements = array(
				'{type}'=>ucfirst($this->_type),
				'{title}'=>($this->_title ? $this->_title : $this->_sourceUrl),
				'{sourceUrl}'=>$this->_sourceUrl,
				'{excerpt}'=>($this->_excerpt ? $this->_excerpt : '+@##@+')
			);
		}
		else {
			$replacements = array(
				'{blogName}'=>($this->_blogName ? $this->_blogName : '+@##@+'),
				'{title}'=>($this->_title ? $this->_title : $this->_sourceUrl),
				'{sourceUrl}'=>$this->_sourceUrl,
			);

		}
		return str_replace(array('\[', '\]'), array('[', ']'),
			preg_replace(
				array(
					'/(?<!\\\\)\[[^\]]*(\+@##@\+|\\\\\])[^\]]*\]/',
					'/(?<!\\\\)\[((?:\\\\\]|[^\]])*)\]/'
				),
				array('', '\1'),
				strtr($this->$template, $replacements)
			)
		);
	}

	/**
	* Sends a pingback or trackback response.
	* @param integer Response code
	* @param array Parameters for response message
	*/
	private function response($code, $params = array()) {
		$msg = Yii::t('linkback', $this->_responses[$code], $params);

		Yii::log(
			"$code: $msg",
			($code ? CLogger::LEVEL_ERROR : CLogger::LEVEL_INFO),
			'linkback.LinkbackController'
		);

		$responseMethod = $this->_type.'Response';

		$response = str_replace('{encoding}', $this->encoding,
				self::XML_HEADER) . $this->$responseMethod($code, $msg);
		header('Content-Length: ' . strlen($response));
		header('Content-Type: text/xml');
		echo $response;
		Yii::app()->end();
	}

	/**
	* Returns an xmlrpcresp object for pingbacks.
	* @param integer Response code
	* @param string Response message
	* @return xmlrpcresp Response object
	*/
	private function pingbackResponse($code, $msg) {
		if ($code) {
			$fc = strtr(self::XML_RPC_MEMBER, array(
					'{name}'=>'faultCode',
					'{value}'=>str_replace('{int}', $code, self::XML_RPC_INT)
			));
			$fs = strtr(self::XML_RPC_MEMBER, array(
					'{name}'=>'faultString',
					'{value}'=>str_replace('{string}', $msg, self::XML_RPC_STRING)
			));
			$response = str_replace('{members}', $fc . $fs, self::XML_RPC_FAULT);
		}
		else {
			$value = str_replace('{string}', $msg, self::XML_RPC_STRING);
			$param = str_replace('{value}', $value, self::XML_RPC_PARAM);
			$response = str_replace('{params}', $param, self::XML_RPC_PARAMS);
		}
		return str_replace('{response}', $response, self::XML_RPC_METHOD_RESPONSE);
	}

	/**
	* Sends a response for trackbacks.
	* @param integer Response code
	* @param string Response message
	* @return boolean true
	*/
	private function trackbackResponse($code, $msg) {
		$response = strtr(self::TRACKBACK_RESPONSE, array(
			'{code}'=>$code,
			'{msg}'=>($code?$msg:'')
		));
		$response = str_replace('<message></message>', '', $response);
		return $response;
	}

	/**
	* Converts a delimited list into an array.
	* If the list is an array it is returned unaltered.
	* @param mixed The list
	* @param string The list delimiter
	* @return array The list converted to an array
	*/
	private function toArray($list,$delimiter=',') {
		return (is_string($list)?array_map('trim',explode($delimiter,$list)):$list);
	}

	/**
	 * This method is invoked after a linkback has been successfully registered.
	 * The default implementation raises the {@link onLinkbackRegistered} event.
	 * Override this method to do postprocessing after linkback registration.
	 * Make sure you call the parent implementation so that the event is raised properly.
	 * @param CModel The target post
	 */
	protected function linkbackRegistered($target) {
		if ($this->hasEventHandler('onLinkbackRegistered'))
			$this->onLinkbackRegistered(new CEvent($this, array(
				'name'=>'onLinkbackRegistered',
				'target'=>$target
			)));
	}

	/**
	 * This event is raised after a linkback has been successfully registered.
	 * @param CEvent The event parameter
	 */
	public function onLinkbackRegistered($event)	{
		$this->raiseEvent('onLinkbackRegistered',$event);
	}
}
